require(data.table)
require(bigalgebra)
require(foreach)
require(doParallel)

# funcao que treina a base
trainingCF <- function(normalized_input_set, lambda, nf) {
  
  # funcao que resolve U, dado I
  solveU <- function(normalized_input_set, lambda, nf, num.users) {
    
    cl<-makeCluster(16, outfile="")
    registerDoParallel(cl)
    
    x <- foreach(i = 1:num.users, .packages=c("data.table", "bigalgebra"), .combine = rbind, .multicombine=T) %dopar% {
      items <- normalized_input_set[user_id==i]$item_id
      ratings <- normalized_input_set[user_id==i]$num_items
      items.feature = attach.big.matrix("items.feature.desc")
      
      nui <- length(items)
      if(nui == 1) {
        Aj <- tcrossprod(items.feature[items]) + lambda*nui*diag(nf)
        Vj <- items.feature[items]*ratings
      } else {
        Aj <- crossprod(items.feature[items]) + lambda*nui*diag(nf)
        Vj <- t(items.feature[items])%*%ratings
      }
      
      t(chol2inv(chol(Aj))%*%Vj)
    }
    
    stopCluster(cl)
    
    temp <- as.big.matrix(x, backingfile = "user.feature.bin", descriptorfile = "user.feature.desc")
  }
  
  
  # funcao que resolve I, dado U
  solveI <- function(normalized_input_set, lambda, nf, num.items) {
    
    cl<-makeCluster(16, outfile="")
    registerDoParallel(cl)
    
    x <- foreach(i = 1:num.items, .packages=c("data.table", "bigalgebra"), .combine = rbind, .multicombine=T) %dopar% {
      users <- normalized_input_set[item_id==i]$user_id
      ratings <- normalized_input_set[item_id==i]$num_items
      user.feature = attach.big.matrix("user.feature.desc")
      
      nui <- length(users)
      if(nui == 1) {
        Aj <- tcrossprod(user.feature[users]) + lambda*nui*diag(nf)
        Vj <- user.feature[users]*ratings
      } else {
        Aj <- crossprod(user.feature[users]) + lambda*nui*diag(nf)
        Vj <- t(user.feature[users])%*%ratings
      }
      
      t(chol2inv(chol(Aj))%*%Vj)
    }
    
    stopCluster(cl)
    
    temp <- as.big.matrix(x, backingfile = "items.feature.bin", descriptorfile = "items.feature.desc")
  }
  
  
  # funcao que calcula RMSE (Raiz Quadrada do Erro Quadratico Medio)
  RMSECalc <- function(normalized_input_set, num.items) {
    
    cl<-makeCluster(16, outfile="")
    registerDoParallel(cl)
    
    y <- foreach(i = 1:num.items, .packages=c("data.table", "bigalgebra"), .combine=sum) %dopar% {
      users <- normalized_input_set[item_id==i]$user_id
      ratings <- normalized_input_set[item_id==i]$num_items
      
      items.feature = attach.big.matrix("items.feature.desc")
      user.feature = attach.big.matrix("user.feature.desc")
      
      
      
      (user.feature[users,] %*% items.feature[i,] - ratings)^2
    }
    
    stopCluster(cl)
    
    return(sqrt(y/num.items))
  }
  
  num.items <- length(unique(normalized_input_set$item_id))
  num.users <- length(unique(normalized_input_set$user_id))
  
  user.feature <- big.matrix(nrow=num.users, ncol=nf, backingfile = "user.feature.bin", descriptorfile = "user.feature.desc")
  items.feature <- matrix(runif(num.items*nf),ncol=nf)
  setkey(normalized_input_set, item_id)
  items.feature[,1] <- normalized_input_set[,list(rate=mean(num_items)),by=item_id]$rate
  items.feature <- as.big.matrix(items.feature, backingfile = "items.feature.bin", descriptorfile = "items.feature.desc")
  
  error <- Inf
  diff <- Inf
  
  # Loop que itera ate atingir o criterio de parada
  while(diff >= 0.0001) {
    solveU(normalized_input_set, lambda, nf, num.users)
    solveI(normalized_input_set, lambda, nf, num.items)
    
    error_ <- RMSECalc(normalized_input_set, num.items)
    diff <- error - error_
    error <- error_
  }

}


# Funcao que retorna as recomendacoes
getRecommendation <- function(user_id, num_recommendation) {
  
  items.feature = attach.big.matrix("items.feature.desc")
  user.feature = attach.big.matrix("user.feature.desc")
  
  all.prevs <- colSums(user.feature[user_id]*t(items.feature[]))
  
  prevs <- which(all.prevs >= sort(all.prevs, decreasing=TRUE)[num_recommendation])
  prevs <- data.frame(cbind(prevs, val=all.prevs[prevs]))
  prevs <- prevs[order(prevs$val, decreasing=TRUE),]
  
  return(prevs$prevs)
}
