// [[Rcpp::depends(BH)]]
#include <Rcpp.h>
using namespace Rcpp;

// [[Rcpp::depends(BH, bigmemory)]]
#include <bigmemory/MatrixAccessor.hpp>

#include <numeric>

// Logic for BigColSums.
template <typename T>
NumericVector BigColSums(XPtr<BigMatrix> pMat, MatrixAccessor<T> mat) {

    // Create the vector we'll store the column sums in.
    NumericVector colSums(pMat->ncol());
    for (size_t i=0; i < pMat->ncol(); ++i)
        colSums[i] = std::accumulate(mat[i], mat[i]+pMat->nrow(), 0.0);
    return colSums;
}


// [[Rcpp::export]]
NumericVector BigColSums(SEXP pBigMat) {
    XPtr<BigMatrix> xpMat(pBigMat);

    switch(xpMat->matrix_type()) {
      case 1:
        return BigColSums(xpMat, MatrixAccessor<char>(*xpMat));
      case 2:
        return BigColSums(xpMat, MatrixAccessor<short>(*xpMat));
      case 4:
        return BigColSums(xpMat, MatrixAccessor<int>(*xpMat));
      case 8:
        return BigColSums(xpMat, MatrixAccessor<double>(*xpMat));
      default:
        throw Rcpp::exception("unknown type detected for big.matrix object!");
    }
}
