require(data.table)
require(doParallel)

buying_intention <- function(orders.training) {

  ## Input de dados: orders.categories (ordens de compra dos clientes), abandoned.cart (produtos abandonados no carrinho) e visits.sku (produtos visitados)
  orders.training <-fread(input="in/base_orders.csv", header=TRUE, sep=";")  
  
  abandoned.cart <- fread(input="in/base_abandoned_cart.csv", header=TRUE, sep=";")
  visits.sku <- fread(input="in/base_surf_history.csv", header=TRUE, sep=";")  
  
  ## Tratamento dos dados: conversão dos campos de data (que estão como string) para classe Date
  orders.training$date <- as.Date(orders.training$date, "%d/%m/%Y")
  abandoned.cart$id_customer <- as.integer(abandoned.cart$id_customer)
  abandoned.cart$qty_abandoned <- as.integer(abandoned.cart$qty_abandoned)
  abandoned.cart$date <- as.Date(abandoned.cart$date, "%d/%m/%Y")
  visits.sku$id_customer <- as.integer(visits.sku$id_customer)
  visits.sku$visits <- as.integer(visits.sku$visits)
  visits.sku$date <- as.Date(visits.sku$date, "%d/%m/%Y")
  
  setkey(abandoned.cart, id_customer)
  setkey(orders.training, id_customer)
  abandoned.cart <- merge(abandoned.cart, unique(orders.training[, c("id_customer"), with=FALSE]))
  
  setkey(visits.sku, id_customer)
  visits.sku <- merge(visits.sku, unique(orders.training[, c("id_customer"), with=FALSE]))
  
  abandoned.cart <- merge(abandoned.cart, unique(orders.training))
  abandoned.cart <- abandoned.cart[,c("id_customer", "date.x", "sku", "qty_abandoned"), with=FALSE]
  setnames(abandoned.cart, "date.x", "date")
  
  visits.sku <- merge(visits.sku, unique(orders.training))
  visits.sku <- visits.sku[,c("id_customer", "date.x", "sku", "visits"), with=FALSE]
  setnames(visits.sku, "date.x", "date")
  
  ## Ordenação e numeração das ordens por cliente
  orders.training[, seq:=order(date, order_nr), by=id_customer]
  setkey(orders.training, id_customer, seq)
  
  ## Cruzamento das ordens, utilizando o id_customer (id do cliente) e a sequencia da ordem
  ## É cruzado com a ordem anterior, retornando date_previous (data anterior) e dessa forma, tem-se o período entre as compras
  orders.join <- orders.training
  orders.join$seq <- orders.join$seq + 1
  setkey(orders.join, id_customer, seq)
  orders.join$order_nr <- NULL
  orders.training <- merge(orders.training, orders.join, all.x=TRUE)
  setnames(orders.training, "date.x", "date")
  setnames(orders.training, "date.y", "date_previous")
  
  ## Caso date == date_previous, utiliza a última data anterior a data da ordem referente.
  equal.dates <- which(orders.training$date==orders.training$date_previous)
  while(length(equal.dates) > 0) {
    orders.join$seq <- orders.join$seq + 1
    setkey(orders.join, id_customer, seq)
    orders.training <- merge(orders.training, orders.join, all.x=TRUE)
    orders.training[equal.dates]$date_previous <- orders.training[equal.dates]$date.y
    setnames(orders.training, "date.x", "date")
    orders.training$date.y <- NULL
    equal.dates <- which(orders.training$date==orders.training$date_previous)
  }
  
  
  ## Atribuição dos produtos abandonados no carrinho as ordens posteriores
  cl<-makeCluster(16, outfile="")
  registerDoParallel(cl)

  orders.abandoned.all <- foreach(i = 1:nrow(orders.training), .packages=c("data.table"), .combine = rbind, .multicombine=T) %dopar% {
    
    if(is.na(orders.training$date_previous[i])) {
      order.abandoned <- abandoned.cart[id_customer==orders.training$id_customer[i] & date <= orders.training$date[i]]
    } else {
      order.abandoned <- abandoned.cart[id_customer==orders.training$id_customer[i] & date <= orders.training$date[i] & date > orders.training$date_previous[i]]
    }
    
    order.abandoned <- cbind(order.abandoned, order_nr=rep(orders$order_nr[i], times=nrow(order.abandoned)))
  }
  stopCluster(cl)
  
  ## Atribuição das visitas nos produtos as ordens posteriores
  cl<-makeCluster(16, outfile="")
  registerDoParallel(cl)
  
  orders.visits.all <- foreach(i = 1:nrow(orders.training), .packages=c("data.table"), .combine = rbind, .multicombine=T) %dopar% {
    
    if(is.na(orders.training$date_previous[i])) {
      order.visits <- visits.sku[id_customer==orders.training$id_customer[i] & date <= orders.training$date[i]]
    } else {
      order.visits <- visits.sku[id_customer==orders.training$id_customer[i] & date <= orders.training$date[i] & date > orders.training$date_previous[i]]
    }
    
    order.visits <- cbind(order.visits, order_nr=rep(orders$order_nr[i], times=nrow(order.visits)))
  }
  stopCluster(cl)
  
  
  setkey(orders.training, id_customer)
  customers.date <- orders.training[,list(date=max(date)), by=id_customer]
  
  visits.no.orders <- data.table()
  abandon.no.orders <- data.table()
  for(i in 1:nrow(customers.date)) {
    visits.no.orders <- rbindlist(list(visits.no.orders, visits.sku[id_customer==customers.date$id_customer[i] & date>customers.date$date[i],c("id_customer", "sku", "visits"), with=FALSE]))
    abandon.no.orders <- rbindlist(list(abandon.no.orders, abandoned.cart[id_customer==customers.date$id_customer[i] & date>customers.date$date[i],c("id_customer", "sku", "qty_abandoned"), with=FALSE]))
  }
  
  visits.no.orders[,buying_intention:=visits/max(visits)*0.5, by=c("sku")]
  abandon.no.orders[, buying_intention:=sum(qty_abandoned), by=c("id_customer", "sku")]
  buying_intention.final <- rbindlist(list(visits.no.orders[,c("id_customer", "sku", "buying_intention"),with=FALSE], abandon.no.orders[,c("id_customer", "sku", "buying_intention"),with=FALSE]))
  buying_intention.final <- buying_intention.final[,list(buying_intention=sum(buying_intention)), by=c("id_customer","sku")]
  write.table(x = buying_intention.final, file = "in/test_set_buying.csv", row.names = FALSE, col.names = TRUE, sep = ";")
  
  ## Normalização dos valores
  orders.visits.all[,buying_intention:=visits/max(visits)*0.5, by=c("sku")]
  orders.abandoned.all[, buying_intention:=sum(qty_abandoned), by=c("order_nr", "sku")]
  
  orders.final <- rbindlist(list(orders.visits.all[,c("order_nr", "sku", "buying_intention"),with=FALSE], orders.abandoned.all[,c("order_nr", "sku", "buying_intention"),with=FALSE]))
  orders.final <- orders.final[,list(buying_intention=sum(buying_intention)), by=c("order_nr","sku")]
  
  return (orders.final)
}
