require(data.table)
require(stringi)
require(stringr)

getFacebookData <- function() {
  
  source("fb_api.R")
  
  # Criacao do arquivo de autenticacao para acesso ao Facebook
  #fb_oauth <- fbOAuth(app_id="810048365689446", app_secret="15e7ad99a1bda244002b0287adc875e6")
  #save(fb_oauth, file = "fb_oauth")
  load("fb_oauth")
  
  # Extracao de dados de posts
  posts <- data.table(getPosts("me", token=fb_oauth, limit=10000))
  
  # tratamento dos dados de posts, para remover os acentos e deixar tudo em letra minuscula
  for(i in 1:nrow(posts)) {
    posts$message_fixed[i] <- gsub("[[:punct:]]", "", tolower(iconv(x=posts$message[i], "WINDOWS-1252","UTF-8")), perl=TRUE)
    posts$message_fixed[i] <- gsub("\n", " ", toString(posts$message_fixed[i]))
  }
  
  # Remover dados anteriores a 2013
  posts$message <- NULL
  posts$created.time <- as.Date(posts$created.time)
  posts <- posts[created.time>="2013-01-01"]
  
  categories <- data.table(read.csv(file="in/item_categories.csv", header=TRUE, sep=";", encoding="UTF-8"))
  
  tagMatrix <- data.table()
  likes.table <- data.table()
  users <- data.table()
  
  # Loop para extrair os likes de cada post e tambem deixar somente as categorias encontradas em cada post, removendo o restante dos textos
  for(i in 1:nrow(posts)) {
    mainWords <- stri_detect_fixed(posts$message_fixed[i], tolower(categories$category_otb))
    
    if(sum(mainWords) == 0) {
      next
    }
    
    likes <- as.data.table(getPostsLikes(post_id=posts$post.id[i], token=fb_oauth))
    likes <- cbind(likes, post.id=rep(posts$post.id[i], times=nrow(likes)))
    
    likes.table <- rbindlist(list(likes.table, likes[,c("likes.id", "post.id"), with=FALSE]))
    users <- rbindlist(list(users, likes[,c("likes.id", "likes.name"), with=FALSE]))
    users <- unique(users)
    
    tags <- cbind(as.data.table(list(tag=tolower(categories$category_otb[which(mainWords==TRUE)]))), post.id=rep(posts$post.id[i], times=sum(mainWords)))
    
    tagMatrix <- rbindlist(list(tagMatrix, unique(tags)))
  }
  
  users$likes.name = tolower(as.character(users$likes.name))
  
  # Leitura dos dados de User da base da loja
  u.customers <- fread(input="in/FM_user.csv", header=TRUE, sep=";")
  u.customers$likes.name <- tolower(paste(str_trim(u.customers$first_name), str_trim(u.customers$last_name), sep=" "))
  u.customers <- u.customers[,c("id_customer", "likes.name"), with=FALSE]
  
  setkey(users, likes.name)
  setkey(u.customers, likes.name)
  
  # Cruzamento dos dados de facebook com os dados da Loja, para conseguir o matching de usuarios
  matching <- merge(users, u.customers)
  matching$likes.id
  
  setnames(matching, "\"likes.id\"", "likes.id")
  matching = matching[,list(likes.id=likes.id[1], id_customer=id_customer[1], count=length(likes.id)), by=likes.name]
  
  setkey(matching, likes.id)
  setkey(likes.table, likes.id)
  
  likes.by.customer <- merge(likes.table, matching[count==1])
  likes.by.customer$likes.name <- NULL
  
  setkey(likes.by.customer, post.id)
  setkey(posts, post.id)
  
  
  # Input dos dados de Ordens
  orders <- fread(input="in/base_orders.csv", header=TRUE, sep=";")
  
  setkey(matching, id_customer)
  setkey(orders, id_customer)
  
  # Cruzamento dos dados de Usuários (matching) com os de Ordens
  orders <- matching[orders]
  orders <- orders[count==1]
  orders <- orders[, c("id_customer", "date", "order_nr"), with=FALSE]
  
  orders$date <- as.Date(orders$date)
  orders <- orders[order(id_customer, date)]
  
  orders[, seq:=order(date, order_nr), by=id_customer]
  setkey(orders, id_customer, seq)
  
  orders.join <- orders
  orders.join$seq <- orders.join$seq + 1
  setkey(orders.join, id_customer, seq)
  orders.join$order_nr <- NULL

  orders <- merge(orders, orders.join, all.x=TRUE)
  setnames(orders, "date.x", "date")
  setnames(orders, "date.y", "date_previous")
  orders$seq <- NULL
  
  equal.dates <- which(orders$date==orders$date_previous)
  for(i in 1:length(equal.dates)) {
    j <- 1
    
    while(!is.na(orders$date_previous[equal.dates[i]]) & orders$date[equal.dates[i]] == orders$date_previous[equal.dates[i]]) {
      orders$date_previous[equal.dates[i]] <- orders$date_previous[equal.dates[i] - j]
      j <- j + 1
    }
  }
  
  likes.by.customer$count <- NULL
  setkey(likes.by.customer, post.id)
  setkey(posts, post.id)
  
  likes.by.customer <- merge(likes.by.customer, posts[,c("post.id", "created.time"), with=FALSE])
  likes.by.customer <- likes.by.customer[order(id_customer, created.time)]
  likes.by.customer$created.time <- as.Date(likes.by.customer$created.time)
  
  orders.likes.all <- data.table()
  
  # Loop para atribuir cada like em post a alguma ordem realizada na loja
  for(i in 1:nrow(orders)) {
    if(is.na(orders$date_previous[i])) {
      order.likes <- likes.by.customer[id_customer==orders$id_customer[i] & created.time <= orders$date[i]]
    } else {
      order.likes <- likes.by.customer[id_customer==orders$id_customer[i] & created.time <= orders$date[i] & created.time > orders$date_previous[i]]
    }
    
    if(nrow(order.likes)==0) {
      next
    }
    
    order.likes <- cbind(order.likes, order_nr=rep(orders$order_nr[i], times=nrow(order.likes)))
    orders.likes.all <- rbindlist(list(orders.likes.all, order.likes))
  }
  
  setkey(orders.likes.all, post.id)
  setkey(tagMatrix, post.id)
  
  orders.likes.all <- merge(orders.likes.all,tagMatrix, allow.cartesian=TRUE)
  
  
  ### Computando TF-IDF ###
  orders.likes.all[,count_tag:=length(post.id), by=c("id_customer","tag")]
  orders.likes.all[,count_tag_customer:=length(tag), by=c("id_customer")]
  
  orders.likes.all[,count_customer_with_tag:=length(unique(id_customer)), by=c("tag")]
  num_customers_tag <- orders.likes.all[,length(unique(id_customer))]
  
  orders.likes.all$tf_idf<-(orders.likes.all$count_tag/orders.likes.all$count_tag_customer)*log10(num_customers_tag/orders.likes.all$count_customer_with_tag)
  
  orders.likes.all <- orders.likes.all[,c("order_nr", "tag", "tf_idf"),with=FALSE]
  
  return (orders.likes.all)
}
