require(data.table)
require(bigalgebra)
require(foreach)
require(doParallel)
require(Rcpp)
sourceCpp("rowSums.cpp")
sourceCpp("colSums.cpp")

# Retorna o rating dado os inputs
recommendation_Tuples <- function(x, x.idx, w0, W, V) {
  return (w0 + sum(W[x.idx]*x) + 0.5*(sum(V[x.idx]*x)^2 - sum((V[x.idx]*x)^2)))
}

# Função que treina a base, gerando os coeficientes w0, W e V
trainingFM <- function(context.facebook, context.buying_intention, context.order, context.user, context.item, ratings, k, mapper) {
  
  num.variables <- nrow(mapper)
  num.training <- nrow(context.order)
  
  # Inicialização das matrizes w0, W e V
  w0 <- big.matrix(nrow=1, ncol=1, backingfile = "w0.bin", descriptorfile = "w0.desc")
  W <- big.matrix(nrow=num.variables, ncol=1, backingfile = "W.bin", descriptorfile = "W.desc")
  V <- big.matrix(nrow=num.variables, ncol=k, backingfile = "V.bin", descriptorfile = "V.desc")
  
  for(i in 1:k) {
    V[,i] <- rnorm(num.variables, mean = 0, sd = 1) / 100
  }
  
  lambda_w0 <- 0
  lambda_W <- 2
  lambda_V <- 2
  
  # Inicialização as matrizes e e q
  e <- big.matrix(nrow=num.training, ncol=1, backingfile = "e.bin", descriptorfile = "e.desc")
  q <- big.matrix(nrow=num.training, ncol=k, backingfile = "q.bin", descriptorfile = "q.desc")
  
  cl<-makeCluster(20, outfile="")
  registerDoParallel(cl)
  
  # Computar os valores de e e q
  foreach(i = 1:num.training, .packages=c("data.table", "bigalgebra"), .combine = c, .multicombine=T) %dopar% {
    
    w0 <- attach.big.matrix("w0.desc")
    W <- attach.big.matrix("W.desc")
    V <- attach.big.matrix("V.desc")
    e <- attach.big.matrix("e.desc")
    q <- attach.big.matrix("q.desc")
    ratings <- attach.big.matrix("ratings.desc")
    
    x.idx <- na.omit(unlist(context.order[,c("id_customer", "age", "gender", "region", "sku", "active", "color", "category", "repurchase", "coupon", "date"), with=FALSE][i]))
    x <- rep(1, times=length(x.idx))
    x.idx <- c(x.idx, unlist(context.facebook[order_nr==context.order$order_nr[i], c("fb_tag"), with=FALSE]), unlist(context.buying_intention[order_nr==context.order$order_nr[i], c("intention_sku"), with=FALSE]))
    x <- c(x, unlist(context.facebook[order_nr==context.order$order_nr[i], c("tf_idf"), with=FALSE]), unlist(context.buying_intention[order_nr==context.order$order_nr[i], c("buying_intention"), with=FALSE]))
    
    e[i] <- recommendation_Tuples(x, x.idx, w0, W, V) - ratings[i]
    
    for(f in 1:k) {
      q[i,f] <- sum(V[x.idx,f]*x)
    }
    
    1
  }
  
  stopCluster(cl)
  
  # Loop principal, que itera calculando todos os coeficientes
  iter <- 1
  while(iter <= 5) {
    w0 <- attach.big.matrix("w0.desc")
    W <- attach.big.matrix("W.desc")
    V <- attach.big.matrix("V.desc")
    e <- attach.big.matrix("e.desc")
    q <- attach.big.matrix("q.desc")
    ratings <- attach.big.matrix("ratings.desc")
    
    # Calculo do coeficiente w0
    w0_ <- -(BigColSums((e - big.matrix(nrow = length(e), ncol = 1, init = w0[]))@address)/(num.training + lambda_w0))
    e[] <- e[] + rep(w0[] - w0_, times=length(e))
    w0[1] <- w0_
    
    
    cl<-makeCluster(20, outfile="")
    registerDoParallel(cl)
    
    
    # Loop de calculo dos coeficientes de W
    foreach(l = 1:num.variables, .packages=c("data.table", "bigalgebra"), .combine = c, .multicombine=T) %dopar% {
      
      W <- attach.big.matrix("W.desc")
      e <- attach.big.matrix("e.desc")
      
      idx.dt <- switch(mapper[id==l]$category, 
                       "id_customer"=context.order[id_customer==l],
                       "region"=context.order[region==l],
                       "gender"=context.order[gender==l],
                       "age"=context.order[age==l],
                       "sku"=context.order[sku==l],
                       "color"=context.order[color==l],
                       "category"=context.order[category==l],
                       "repurchase"=context.order[repurchase==l],
                       "coupon"=context.order[coupon==l],
                       "active"=context.order[active==l],
                       "date"=context.order[date==l],
                       "intention_sku"=context.buying_intention[intention_sku==l],
                       "fb_tag"=context.facebook[fb_tag==l]
      )
      
      if(mapper[id==l]$category == "intention_sku") {
        idx.dt <- merge(context.order, idx.dt, by="order_nr")
        x.idx <- idx.dt$pos
        x <- idx.dt$buying_intention
      } else if(mapper[id==l]$category == "fb_tag") {
        idx.dt <- merge(context.order, idx.dt, by="order_nr")
        x.idx <- idx.dt$pos
        x <- idx.dt$tf_idf
      } else {
        x.idx <- idx.dt$pos
        x <- rep(1, times=length(x.idx))
      }
      
      W_ <- -sum((e[x.idx] - W[l]*x)*x) / (sum(x^2) + lambda_W)
      e[x.idx] <- e[x.idx] + (W_ - W[l])*x
      W[l] <- W_
      
      1
    }
    
    stopCluster(cl)
    
    
    # Loop externo que itera sobre cada valor de k
    for(f in 1:k) {
      
      #setup parallel backend to use 20 processors
      cl<-makeCluster(20, outfile="")
      registerDoParallel(cl)
      
      # Loop que calcula todos os coeficientes de V
      foreach(l = 1:num.variables, .packages=c("data.table", "bigalgebra"), .combine = c, .multicombine=T) %dopar% {
        
        V <- attach.big.matrix("V.desc")
        e <- attach.big.matrix("e.desc")
        q <- attach.big.matrix("q.desc")
        
        idx.dt <- switch(mapper[id==l]$category, 
                         "id_customer"=context.order[id_customer==l],
                         "region"=context.order[region==l],
                         "gender"=context.order[gender==l],
                         "age"=context.order[age==l],
                         "sku"=context.order[sku==l],
                         "color"=context.order[color==l],
                         "category"=context.order[category==l],
                         "repurchase"=context.order[repurchase==l],
                         "coupon"=context.order[coupon==l],
                         "date"=context.order[date==l],
                         "active"=context.order[active==l],
                         "intention_sku"=context.buying_intention[intention_sku==l],
                         "fb_tag"=context.facebook[fb_tag==l]
        )
        
        if(mapper[id==l]$category == "intention_sku") {
          idx.dt <- merge(context.order, idx.dt, by="order_nr")
          x.idx <- idx.dt$pos
          x <- idx.dt$buying_intention
        } else if(mapper[id==l]$category == "fb_tag") {
          idx.dt <- merge(context.order, idx.dt, by="order_nr")
          x.idx <- idx.dt$pos
          x <- idx.dt$tf_idf
        } else {
          x.idx <- idx.dt$pos
          x <- rep(1, times=length(x.idx))
        }
        
        h <- x * q[x.idx,f] - x^2*V[l,f]
        V_ <- -sum((e[x.idx] - V[l,f]*x)*h) / (sum(h^2) + lambda_V)
        e[x.idx] <- e[x.idx] + (V_ - V[l,f])*x
        q[x.idx,f] <- q[x.idx,f] + (V_ - V[l,f])*x
        V[l,f] <- V_
        
        1
      }
      
      stopCluster(cl)
      
    }
    
    
    iter <- iter + 1
  }
  
}

getRecommendation <- function(context.facebook, context.buying_intention, context.user, context.item, mapper, i) {
  w0 <- attach.big.matrix("w0.desc")
  W <- attach.big.matrix("W.desc")
  V <- attach.big.matrix("V.desc")
  
  x.idx <- na.omit(unlist(context.user[,c("id_customer", "age", "gender", "region"), with=FALSE][i]))
  x <- rep(1, times=length(x.idx))
  x.idx <- c(x.idx, unlist(context.facebook[id_customer==context.user$id_customer[i], c("tag"), with=FALSE]), unlist(context.buying_intention[id_customer==context.user$id_customer[i], c("intention_sku"), with=FALSE]))
  x <- c(x, unlist(context.facebook[id_customer==context.user$id_customer[i], c("tf_idf"), with=FALSE]), unlist(context.buying_intention[id_customer==context.user$id_customer[i], c("buying_intention"), with=FALSE]))
  
  all.prevs <- NA
  for(j in 1:nrow(context.item)) {
    y.idx <- c(x.idx, unlist(context.item[,c("sku", "category", "color"), with=FALSE][j]))
    y <- c(x, rep(1, times=ncol(context.item[,c("sku", "category", "color"), with=FALSE])))
    
    all.prevs[j] <- recommendation_Tuples(y, y.idx, w0, W, V)
  }
  
  prevs <- which(all.prevs >= sort(all.prevs, decreasing=TRUE)[num_recommendation])
  prevs <- data.frame(cbind(prevs, val=all.prevs[prevs]))
  prevs <- prevs[order(prevs$val, decreasing=TRUE),]
  
  return(prevs$prevs)
}
