library(RCurl)
library(RJSONIO)
library(httr)

# funcao que cria e retorna o token de acesso ao Facebook
fbOAuth <- function(app_id, app_secret, extended_permissions=TRUE)
{
  ## getting callback URL
  full_url <- oauth_callback()
  full_url <- gsub("(.*localhost:[0-9]{1,5}/).*", x=full_url, replacement="\\1")
  message <- paste("Copy and paste into Site URL on Facebook App Settings:", full_url, "\nWhen done, press any key to continue...")
  ## prompting user to introduce callback URL in app page
  invisible(readline(message))
  ## a simplified version of the example in httr package
  facebook <- oauth_endpoint( authorize = "https://www.facebook.com/dialog/oauth", access = "https://graph.facebook.com/oauth/access_token")  
  myapp <- oauth_app("facebook", app_id, app_secret)
  if (extended_permissions==TRUE){
    scope <- paste("user_birthday,user_hometown,user_location,user_relationships,",
                   "friends_birthday,friends_hometown,friends_location,friends_relationships,publish_actions,",
                   "user_status,user_checkins,friends_status,friends_checkins,user_likes,friends_likes,read_stream,export_stream", 
                   collapse="")
  }
  else { scope <- NULL}
  
  ## before httr 0.3
  if (packageVersion('httr')$minor < 3){
    facebook_token <- oauth2.0_token(facebook, myapp,
                                     scope=scope, type = "application/x-www-form-urlencoded")
    fb_oauth <- sign_oauth2.0(facebook_token$access_token) 
    if (GET("https://graph.facebook.com/me", config=fb_oauth)$status==200){
      message("Authentication successful.")
    }
  }
  
  ## with httr 0.3
  if (packageVersion('httr')$minor >= 3){
    fb_oauth <- oauth2.0_token(facebook, myapp,
                               scope=scope, type = "application/x-www-form-urlencoded", cache=FALSE)	
    if (GET("https://graph.facebook.com/me", config(token=fb_oauth))$status==200){
      message("Authentication successful.")
    }	
  }
  
  return(fb_oauth)
}

# funcao que transforma os dados de posts em Json para data.frame
getJsonPostsDf <- function(data) {
  
  l <- list(
    post.id = lapply(data, '[[', 'id'),
    created.time = lapply(data, '[[', 'created_time'),
    message = lapply(data, '[[', 'message')
  )
  
  # replace all occurances of NULL with NA
  df = data.frame(do.call("cbind", lapply(l, function(x) sapply(x, function(xx) ifelse(is.null(xx), NA, xx)))))
  
  return(df)
}

# funcao que transforma os dados de likes em Json para data.frame
getJsonLikesDf <- function(data) {
  
  l <- list(
    likes.id = lapply(data, '[[', 'id'),
    likes.name = lapply(data, '[[', 'name')
  )
  
  # replace all occurances of NULL with NA
  df = data.frame(do.call("cbind", lapply(l, function(x) sapply(x, function(xx) ifelse(is.null(xx), NA, xx)))))
  
  return(df)
}

# funcao que puxa os dados de posts
getPosts <- function(name, token, limit=100) {
  # Construct Facebook Graph API URL
  u <- paste("https://graph.facebook.com/", name, "/posts?fields=id,message&limit=", limit, sep = "")
  
  # Get data
  json <- GET(u, config(token=token))
  content <- fromJSON(rawToChar(json$content))
  data <- content$data
  data[which(sapply(data,length)!=3)] <- NULL
  
  df.posts <- getJsonPostsDf(data)
  
  while(nrow(df.posts) < limit) {
    u <- content$paging[2]
    
    json <- GET(u, config(token=token))
    content <- fromJSON(rawToChar(json$content))
    data <- content$data
    if(length(data)==0)
      break
    data[which(sapply(data,length)!=3)] <- NULL
    
    df.posts <- rbind(df.posts, getJsonPostsDf(data))
  }
  
  return(df.posts)
}

# funcao que puxa os dados de likes
getPostsLikes <- function(post_id, token, limit=10000) {
  u <- paste("https://graph.facebook.com/", post_id , "?fields=likes.limit(", limit, ").fields(id,name)", sep="")
  
  json <- GET(u, config(token=token))
  content <- fromJSON(rawToChar(json$content))
  data <- content$likes$data
  
  df.likes <- getJsonLikesDf(data)
  
  return (df.likes)
}

