require(data.table)
require(bigalgebra)
options(bigmemory.allow.dimnames=TRUE)

setwd("/home/ruser/Recommender_system/")
source("context_buying_intention.R")
source("factorization_machine.R")

# funcao para treinar a base
train_data <- function() {
  # input de dados do facebook
  context.facebook <- fread(input='in/FM_context_facebook.csv', header=TRUE, sep='\t')
  context.facebook <- unique(context.facebook)
  colnames(context.facebook) <- c("order_nr", "fb_tag", "tf_idf")
  
  # input de dados de intencao de compra
  context.buying_intention <- buying_intention(context.facebook[,c("order_nr"),with=FALSE])
  setnames(context.buying_intention, "sku", "intention_sku")
  
  # input de dados de Order, User e Item
  context.order <- fread(input='in/FM_orders.csv', header=TRUE, sep=';')
  user_in <- fread(input='in/FM_user.csv', header=TRUE, sep=';')
  item_in <- fread(input='in/FM_item.csv', header=TRUE, sep=';')
  
  # tratamento inicial dos dados
  user_in$first_name <- NULL
  user_in$last_name <- NULL
  user_in[gender=="NULL", gender := NA]
  user_in[age=="NULL", age := NA]
  user_in[region=="NULL", region := NA]
  
  item_in[category=="NULL", category := NA]
  item_in[color=="NULL", color := NA]
  item_in[active=="NULL", active := NA]
  
  setkey(context.order, id_customer)
  setkey(user_in, id_customer)
  context.user <- merge(user_in, unique(context.order[,c("id_customer"),with=FALSE]))
  
  setkey(context.order, sku)
  setkey(item_in, sku)
  context.item <- merge(item_in, unique(context.order[,c("sku"),with=FALSE]))
  
  
  mapper <- NULL
  num.ids <- 0
  
  context.user.names <- names(context.user)
  
  # Loop para mapear os valores de User, deixando todos na mesma base
  for(i in 1:length(context.user.names)) {
    temp <- na.omit(unique(context.user[,context.user.names[i],with=FALSE]))
    setkeyv(temp, context.user.names[i])
    temp$id <- seq.int(from = num.ids + 1, to = num.ids+nrow(temp))
    
    num.ids <- num.ids+nrow(temp)
    
    setkeyv(context.user, context.user.names[i])
    context.user <- temp[context.user]
    context.user <- context.user[,!context.user.names[i],with=FALSE]
    
    setnames(context.user, "id", context.user.names[i])
    
    mapper <- rbindlist(list(mapper, cbind(rep("user", times=nrow(temp)), rep(context.user.names[i], times=nrow(temp)), temp)))
  }
  
  
  gc()
  
  
  context.item.names <- names(context.item)
  
  # Loop para mapear os valores de Item, deixando todos na mesma base
  for(i in 1:length(context.item.names)) {
    temp <- na.omit(unique(context.item[,context.item.names[i],with=FALSE]))
    setkeyv(temp, context.item.names[i])
    temp$id <- seq.int(from = num.ids + 1, to = num.ids+nrow(temp))
    
    num.ids <- num.ids+nrow(temp)
    
    setkeyv(context.item, context.item.names[i])
    context.item <- temp[context.item]
    context.item <- context.item[,!context.item.names[i],with=FALSE]
    
    setnames(context.item, "id", context.item.names[i])
    
    mapper <- rbindlist(list(mapper, cbind(rep("item", times=nrow(temp)), rep(context.item.names[i], times=nrow(temp)), temp)))
  }
  
  gc()
  
  
  # Mapear os dados de Facebook
  temp <- na.omit(unique(context.facebook[,"fb_tag",with=FALSE]))
  setkey(temp, fb_tag)
  temp$id <- seq.int(from = num.ids + 1, to = num.ids+nrow(temp))
  
  num.ids <- num.ids+nrow(temp)
  
  setkey(context.facebook, fb_tag)
  context.facebook <- temp[context.facebook]
  context.facebook <- context.facebook[,!"fb_tag",with=FALSE]
  
  setnames(context.facebook, "id", "fb_tag")
  
  mapper <- rbindlist(list(mapper, cbind(rep("context", times=nrow(temp)), rep("fb_tag", times=nrow(temp)), temp)))
  
  
  gc()
  
  
  # Mapear os dados de Intencao de Compra
  temp <- na.omit(unique(context.buying_intention[,"intention_sku",with=FALSE]))
  setkey(temp, "intention_sku")
  temp$id <- seq.int(from = num.ids + 1, to = num.ids+nrow(temp))
  
  num.ids <- num.ids+nrow(temp)
  
  setkey(context.buying_intention, intention_sku)
  context.buying_intention <- temp[context.buying_intention]
  context.buying_intention <- context.buying_intention[,!"intention_sku",with=FALSE]
  setnames(context.buying_intention, "id", "intention_sku")
  mapper <- rbindlist(list(mapper, cbind(rep("context", times=nrow(temp)), rep("intention_sku", times=nrow(temp)), temp)))
  
  
  gc()
  
  
  context.order.names <- names(context.order[,!c("order_nr", "id_customer", "sku", "qty_item"),with=FALSE])
  
  # Loop para mapear os valores de Order, deixando todos na mesma base
  for(i in 1:length(context.order.names)) {
    temp <- na.omit(unique(context.order[,context.order.names[i],with=FALSE]))
    setkeyv(temp, context.order.names[i])
    temp$id <- seq.int(from = num.ids + 1, to = num.ids+nrow(temp))
    
    num.ids <- num.ids+nrow(temp)
    
    setkeyv(context.order, context.order.names[i])
    context.order <- temp[context.order]
    context.order <- context.order[,!context.order.names[i],with=FALSE]
    
    setnames(context.order, "id", context.order.names[i])
    
    mapper <- rbindlist(list(mapper, cbind(rep("context", times=nrow(temp)), rep(context.order.names[i], times=nrow(temp)), temp)))
  }
  
  setnames(mapper, "V1", "type")
  setnames(mapper, "V2", "category")
  setnames(mapper, "id_customer", "value")
  
  
  # Tratamento das bases para deixar todos de acordo com a estrutura de mapper
  setkey(mapper, value)
  context.order$id_customer <- as.character(context.order$id_customer)
  setnames(context.order, "id_customer", "value")
  setkey(context.order, value)
  
  context.order <- merge(context.order, mapper[category=="id_customer", c("value", "id"), with=FALSE])
  context.order <- context.order[, c("order_nr", "id", "sku", "repurchase", "coupon", "date", "qty_item"), with=FALSE]
  setnames(context.order, "id", "id_customer")
  
  
  setnames(context.order, "sku", "value")
  setkey(context.order, value)
  context.order <- merge(context.order, mapper[category=="sku", c("value", "id"), with=FALSE])
  context.order <- context.order[, c("order_nr", "id_customer", "id", "repurchase", "coupon", "date", "qty_item"), with=FALSE]
  setnames(context.order, "id", "sku")
  
  setkey(context.order, sku)
  setkey(context.item, sku)
  context.order <- context.item[context.order]
  
  setkey(context.order, id_customer)
  setkey(context.user, id_customer)
  context.order <- context.user[context.order]
  
  ratings <- as.big.matrix(context.order$qty_item, descriptorfile = "ratings.desc", backingfile = "ratings.bin")
  
  context.order$qty_item <- NULL
  context.order$pos <- seq.int(from = 1, to = nrow(context.order))
  
  rm(context.user)
  rm(context.item)
  rm(user_in)
  rm(item_in)
  rm(temp)
  
  gc()
  
  k <- 5
  
  trainingFM(context.facebook, context.buying_intention, context.order, context.user, context.item, ratings, 5, mapper)
}


context.facebook <- data.table(read.table(file="in/test_set_fb.csv", sep = "\t", header = TRUE, encoding = "UTF-8"))
context.facebook$id_customer <- as.character(context.facebook$id_customer)
context.facebook$tag <- as.character(context.facebook$tag)

context.buying_intention <- fread(input = "in/test_set_buying.csv", sep = ";", header = TRUE)
setnames(context.buying_intention, "sku", "intention_sku")
context.buying_intention$id_customer <- as.character(context.buying_intention$id_customer)
context.buying_intention$intention_sku <- as.character(context.buying_intention$intention_sku)

user_in <- fread(input='in/FM_user.csv', header=TRUE, sep=';')
item_in <- fread(input='in/FM_item.csv', header=TRUE, sep=';')
crm_base <- fread(input = "in/crm_base.csv", header = TRUE)
crm_base$id_customer <- as.character(crm_base$id_customer)

user_in$first_name <- NULL
user_in$last_name <- NULL
user_in[gender=="NULL", gender := NA]
user_in[age=="NULL", age := NA]
user_in[region=="NULL", region := NA]

item_in[category=="NULL", category := NA]
item_in[color=="NULL", color := NA]
item_in[active=="NULL", active := NA]

setkey(context.buying_intention, id_customer)
setkey(context.facebook, id_customer)

setkey(mapper, value)
mapper.filtered <- mapper[category=="id_customer"]
context.buying_intention <- mapper.filtered[context.buying_intention, c("id", "intention_sku", "buying_intention"), with=FALSE]
context.facebook <- mapper.filtered[context.facebook, c("id", "tag", "tf_idf"), with=FALSE]
context.facebook <- context.facebook[!is.na(id)]
setnames(context.buying_intention, "id", "id_customer")
setnames(context.facebook, "id", "id_customer")
crm_base <- mapper.filtered[crm_base]
crm_base <- crm_base[!is.na(id)]
crm_base <- crm_base [, "id", with=FALSE]
setnames(crm_base, "id", "id_customer")

setkey(context.buying_intention, intention_sku)
mapper.filtered <- mapper[category=="intention_sku"]
context.buying_intention <- mapper.filtered[context.buying_intention, c("id_customer", "id", "buying_intention"), with=FALSE]
context.buying_intention <- context.buying_intention[!is.na(id)]
setnames(context.buying_intention, "id", "intention_sku")

setkey(context.facebook, tag)
mapper.filtered <- mapper[category=="fb_tag"]
context.facebook <- mapper.filtered[context.facebook, c("id_customer", "id", "tf_idf"), with=FALSE]
context.facebook <- context.facebook[!is.na(id)]
setnames(context.facebook, "id", "fb_tag")

setkey(crm_base, id_customer)
setkey(context.buying_intention, id_customer)
setkey(context.facebook, id_customer)

context.buying_intention <- merge(context.buying_intention, crm_base)
context.facebook <- merge(context.facebook, crm_base)

setkey(user_in, id_customer)

crm_base$id_customer <- as.integer(crm_base$id_customer)
crm_base <- merge(user_in, crm_base)

for(i in 1:nrow(crm_base)) {
  recommendations[i] <- getRecommendation(context.facebook, context.buying_intention, crm_base, item_in, mapper, i)
}

